/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#ifndef __gulm_config_h__
#define __gulm_config_h__
#include "LLi.h"

/* I may want to add some kind of marker to every option that this holds.
 * A couple of bits to store weither or not items are different than the
 * default, and possibly other info.  Mostly if I work up the live config
 * modification changes that others have requested, it would be nice to be
 * able to dump a live config that only contained what was needed, and not
 * everything.
 *
 * I may be able to do this just by `diffing' against the defaults too.
 *
 * I should take the time and evaluate how much of this really truely needs
 * to be tightly tied into gulm.
 */

typedef struct {
   LLi_t in_list;
   uint32_t ip;
   uint8_t *name;
} ip_name_t;

typedef struct {
   uint32_t hashval;/* hash of the config. */
   uint8_t *clusterID;
   uint8_t *fencebin;
   uint8_t *run_as;
   uint8_t *lock_file;

   uint16_t corePort;
   uint64_t heartbeat_rate; /* milli-sec res, ignore the micro-secs. */
   uint64_t master_scan_delay;
   uint64_t new_con_timeout;
   uint16_t allowed_misses;

   /* I want to fix this var to hold the finalized quorum values instead of
    * the mess it has right now.
    */
   uint16_t quorum; /* >node_cnt: auto, =0:none, >0:usethis */
   uint16_t use_quorum; /* what is the actual calculated quorum value? */

   uint16_t fog; /* true || false */

   enum {nvt_tcpwrapper, nvt_ccs, nvt_allow} node_verification_type;

   uint16_t node_cnt;
   LLi_t node_list;

   uint16_t how_many_lts; /* int */
   uint16_t lt_port;
   uint16_t ltpx_port;

   unsigned int lt_cf_rate;
   unsigned long lt_maxlocks;
   unsigned int lt_hashbuckets;
   unsigned int lt_prelocks;
   unsigned int lt_prelkrqs;
   unsigned int lt_preholds;

   uint64_t cb_recheck_rate; /* milli-sec res, ignore the micro-secs. */
} gulm_config_t;


/* prototypes */
void release_gulm_config(gulm_config_t *gf);
void free_gulm_config(gulm_config_t *gf);
int rebuild_server_list(gulm_config_t *gf);
int parse_conf(gulm_config_t *gf);
void dump_conf(gulm_config_t *gf);
int serialize_config(gulm_config_t *gf, int fd);
int Can_I_be_a_master(gulm_config_t *gf, uint32_t ip);
int get_lt_range(int which, int of, int *start, int *stop);
int get_ip_for_name_ccs_libc(char *name, uint32_t *ip);

#define get_name_for_ip get_name_for_ip_libc
#define get_ip_for_name get_ip_for_name_ccs_libc

#endif /*__gulm_config_h__*/
/* vim: set ai cin et sw=3 ts=3 : */
