/******************************************************************************
*******************************************************************************
**
**  Copyright (C) 2006 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#ifndef __gulm_log_h__
#define __gulm_log_h__

/* do not access these variables directly */
extern int log_running;
extern char *syslog_msg;
extern unsigned int syslog_max_size;

#include "gulm_log_msg_bits.h"

/* no syslog if debugging. */
#ifndef DEBUG
#include <syslog.h>

#define log(v, fmt, args...) do { \
	if (log_running) { \
		snprintf(syslog_msg, syslog_max_size, fmt , ## args); \
		log_to_buffer(v); \
	} \
	else \
		syslog(v, fmt , ## args); \
} while(0)

#define log_msg(v, fmt, args...) if(((v)&verbosity)==(v)||(v)==lgm_Always) {\
   log(LOG_NOTICE, fmt , ## args );}
#define log_err(fmt, args...){ \
   log(LOG_ERR, "ERROR [%s:%d] " fmt , __FILE__ , __LINE__ , ## args ); }
#define log_bug(fmt, args...) { \
   log(LOG_NOTICE, "BUG[%s:%d] " fmt , __FILE__ , __LINE__ , ## args ); }
#define die(ext, fmt, args...) { \
   fprintf(stderr, "In %s:%d (%s) death by:\n" fmt , __FILE__ , \
         __LINE__ , GULM_RELEASE_NAME , ## args ); \
   log(LOG_ERR, "In %s:%d (%s) death by:\n" fmt , __FILE__ , \
         __LINE__ , GULM_RELEASE_NAME , ## args ); exit(ext);}
#define TICK log_bug("TICK.\n")
#else /*DEBUG*/
#define log_msg(v, fmt, args...) if(((v)&verbosity)==(v)||(v)==lgm_Always){ \
   fprintf(stdout, "%s: " fmt , ProgramName , ## args ); }
#define log_err(fmt, args...) {\
   fprintf(stderr, "ERROR [%s:%s:%d] " fmt , ProgramName , __FILE__ , \
         __LINE__ , ## args ); }
#define log_bug(fmt, args...) {\
   fprintf(stderr, "[bug:%s:%s:%d] " fmt , ProgramName , __FILE__ , \
         __LINE__ , ## args ); }
#define die(ext, fmt, args...) {\
   fprintf(stderr, "In %s:%d (%s) death by:\n" fmt , __FILE__ , \
         __LINE__ , GULM_RELEASE_NAME , ## args ); exit(ext);}
#define TICK log_bug("TICK.\n")
#endif /*DEBUG*/


void log_init(unsigned int log_size, unsigned int max_size);

/* do not access this function directly */
void log_to_buffer(int priority);

#endif /* __gulm_log_h__ */
